package java.time;

import java.time.chrono.ChronoLocalDate;
import java.time.chrono.ChronoZonedDateTime;
import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.time.temporal.TemporalQueries;
import java.time.temporal.TemporalQuery;
import java.time.temporal.ChronoUnit;
import java.time.temporal.TemporalAmount;
import java.time.temporal.ValueRange;
import java.time.temporal.TemporalUnit;
import java.time.temporal.TemporalField;
import java.time.format.DateTimeFormatter;
import java.time.temporal.TemporalAccessor;
import java.time.temporal.ChronoField;
import java.util.Objects;
import java.io.Serializable;
import java.time.chrono.ChronoLocalDateTime;
import java.time.temporal.TemporalAdjuster;
import java.time.temporal.Temporal;

public final class LocalDateTime implements Temporal, TemporalAdjuster, ChronoLocalDateTime<LocalDate>, Serializable
{
    public static final LocalDateTime MIN;
    public static final LocalDateTime MAX;
    private static final long serialVersionUID = 6207766400415563566L;
    private final LocalDate date;
    private final LocalTime time;
    
    public static LocalDateTime now() {
        return now(Clock.systemDefaultZone());
    }
    
    public static LocalDateTime now(ZoneId zoneId) {
        return now(Clock.system(zoneId));
    }
    
    public static LocalDateTime now(Clock obj) {
        Objects.requireNonNull(obj, "clock");
        Instant instant = obj.instant();
        return ofEpochSecond(instant.getEpochSecond(), instant.getNano(), obj.getZone().getRules().getOffset(instant));
    }
    
    public static LocalDateTime of(int n, Month month, int n2, int n3, int n4) {
        return new LocalDateTime(LocalDate.of(n, month, n2), LocalTime.of(n3, n4));
    }
    
    public static LocalDateTime of(int n, Month month, int n2, int n3, int n4, int n5) {
        return new LocalDateTime(LocalDate.of(n, month, n2), LocalTime.of(n3, n4, n5));
    }
    
    public static LocalDateTime of(int n, Month month, int n2, int n3, int n4, int n5, int n6) {
        return new LocalDateTime(LocalDate.of(n, month, n2), LocalTime.of(n3, n4, n5, n6));
    }
    
    public static LocalDateTime of(int n, int n2, int n3, int n4, int n5) {
        return new LocalDateTime(LocalDate.of(n, n2, n3), LocalTime.of(n4, n5));
    }
    
    public static LocalDateTime of(int n, int n2, int n3, int n4, int n5, int n6) {
        return new LocalDateTime(LocalDate.of(n, n2, n3), LocalTime.of(n4, n5, n6));
    }
    
    public static LocalDateTime of(int n, int n2, int n3, int n4, int n5, int n6, int n7) {
        return new LocalDateTime(LocalDate.of(n, n2, n3), LocalTime.of(n4, n5, n6, n7));
    }
    
    public static LocalDateTime of(LocalDate obj, LocalTime obj2) {
        Objects.requireNonNull(obj, "date");
        Objects.requireNonNull(obj2, "time");
        return new LocalDateTime(obj, obj2);
    }
    
    public static LocalDateTime ofInstant(Instant instant, ZoneId obj) {
        Objects.requireNonNull(instant, "instant");
        Objects.requireNonNull(obj, "zone");
        return ofEpochSecond(instant.getEpochSecond(), instant.getNano(), obj.getRules().getOffset(instant));
    }
    
    public static LocalDateTime ofEpochSecond(long n, int n2, ZoneOffset obj) {
        Objects.requireNonNull(obj, "offset");
        ChronoField.NANO_OF_SECOND.checkValidValue((long)n2);
        long n3 = n + (long)obj.getTotalSeconds();
        return new LocalDateTime(LocalDate.ofEpochDay(Math.floorDiv(n3, 86400L)), LocalTime.ofNanoOfDay((long)(int)Math.floorMod(n3, 86400L) * 1000000000L + (long)n2));
    }
    
    public static LocalDateTime from(TemporalAccessor obj) {
        if ((obj instanceof LocalDateTime)) {
            return (LocalDateTime)obj;
        }
        if ((obj instanceof ZonedDateTime)) {
            return ((ZonedDateTime)obj).toLocalDateTime();
        }
        if ((obj instanceof OffsetDateTime)) {
            return ((OffsetDateTime)obj).toLocalDateTime();
        }
        try {
            return new LocalDateTime(LocalDate.from(obj), LocalTime.from(obj));
        }
        catch (DateTimeException ex) {
            throw new DateTimeException("Unable to obtain LocalDateTime from TemporalAccessor: " + obj + " of type " + obj.getClass().getName(), ex);
        }
    }
    
    public static LocalDateTime parse(CharSequence charSequence) {
        return parse(charSequence, DateTimeFormatter.ISO_LOCAL_DATE_TIME);
    }
    
    public static LocalDateTime parse(CharSequence text, DateTimeFormatter obj) {
        Objects.requireNonNull(obj, "formatter");
        return (LocalDateTime)obj.parse(text, LocalDateTime::from);
    }
    
    private LocalDateTime(LocalDate date, LocalTime time) {
        this.date = date;
        this.time = time;
    }
    
    private LocalDateTime with(LocalDate localDate, LocalTime localTime) {
        if (this.date == localDate && this.time == localTime) {
            return this;
        }
        return new LocalDateTime(localDate, localTime);
    }
    
    @Override
    public boolean isSupported(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            ChronoField chronoField = (ChronoField)temporalField;
            return chronoField.isDateBased() || chronoField.isTimeBased();
        }
        return temporalField != null && temporalField.isSupportedBy(this);
    }
    
    @Override
    public boolean isSupported(TemporalUnit unit) {
        return super.isSupported(unit);
    }
    
    @Override
    public ValueRange range(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            return ((ChronoField)temporalField).isTimeBased() ? this.time.range(temporalField) : this.date.range(temporalField);
        }
        return temporalField.rangeRefinedBy(this);
    }
    
    @Override
    public int get(TemporalField field) {
        if ((field instanceof ChronoField)) {
            return ((ChronoField)field).isTimeBased() ? this.time.get(field) : this.date.get(field);
        }
        return super.get(field);
    }
    
    @Override
    public long getLong(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            return ((ChronoField)temporalField).isTimeBased() ? this.time.getLong(temporalField) : this.date.getLong(temporalField);
        }
        return temporalField.getFrom(this);
    }
    
    @Override
    public LocalDate toLocalDate() {
        return this.date;
    }
    
    public int getYear() {
        return this.date.getYear();
    }
    
    public int getMonthValue() {
        return this.date.getMonthValue();
    }
    
    public Month getMonth() {
        return this.date.getMonth();
    }
    
    public int getDayOfMonth() {
        return this.date.getDayOfMonth();
    }
    
    public int getDayOfYear() {
        return this.date.getDayOfYear();
    }
    
    public DayOfWeek getDayOfWeek() {
        return this.date.getDayOfWeek();
    }
    
    @Override
    public LocalTime toLocalTime() {
        return this.time;
    }
    
    public int getHour() {
        return this.time.getHour();
    }
    
    public int getMinute() {
        return this.time.getMinute();
    }
    
    public int getSecond() {
        return this.time.getSecond();
    }
    
    public int getNano() {
        return this.time.getNano();
    }
    
    @Override
    public LocalDateTime with(TemporalAdjuster temporalAdjuster) {
        if ((temporalAdjuster instanceof LocalDate)) {
            return this.with((LocalDate)temporalAdjuster, this.time);
        }
        if ((temporalAdjuster instanceof LocalTime)) {
            return this.with(this.date, (LocalTime)temporalAdjuster);
        }
        if ((temporalAdjuster instanceof LocalDateTime)) {
            return (LocalDateTime)temporalAdjuster;
        }
        return (LocalDateTime)temporalAdjuster.adjustInto(this);
    }
    
    @Override
    public LocalDateTime with(TemporalField temporalField, long n) {
        if ((temporalField instanceof ChronoField)) {
            if (((ChronoField)temporalField).isTimeBased()) {
                return this.with(this.date, this.time.with(temporalField, n));
            }
            return this.with(this.date.with(temporalField, n), this.time);
        }
        else {
            return (LocalDateTime)temporalField.adjustInto(this, n);
        }
    }
    
    public LocalDateTime withYear(int n) {
        return this.with(this.date.withYear(n), this.time);
    }
    
    public LocalDateTime withMonth(int n) {
        return this.with(this.date.withMonth(n), this.time);
    }
    
    public LocalDateTime withDayOfMonth(int n) {
        return this.with(this.date.withDayOfMonth(n), this.time);
    }
    
    public LocalDateTime withDayOfYear(int n) {
        return this.with(this.date.withDayOfYear(n), this.time);
    }
    
    public LocalDateTime withHour(int n) {
        return this.with(this.date, this.time.withHour(n));
    }
    
    public LocalDateTime withMinute(int n) {
        return this.with(this.date, this.time.withMinute(n));
    }
    
    public LocalDateTime withSecond(int n) {
        return this.with(this.date, this.time.withSecond(n));
    }
    
    public LocalDateTime withNano(int n) {
        return this.with(this.date, this.time.withNano(n));
    }
    
    public LocalDateTime truncatedTo(TemporalUnit temporalUnit) {
        return this.with(this.date, this.time.truncatedTo(temporalUnit));
    }
    
    @Override
    public LocalDateTime plus(TemporalAmount obj) {
        if ((obj instanceof Period)) {
            return this.with(this.date.plus((TemporalAmount)obj), this.time);
        }
        Objects.requireNonNull(obj, "amountToAdd");
        return (LocalDateTime)obj.addTo(this);
    }
    
    @Override
    public LocalDateTime plus(long n, TemporalUnit temporalUnit) {
        if ((temporalUnit instanceof ChronoUnit)) {
            switch ((ChronoUnit)temporalUnit) {
                case NANOS: {
                    return this.plusNanos(n);
                }
                case MICROS: {
                    return this.plusDays(n / 86400000000L).plusNanos(n % 86400000000L * 1000L);
                }
                case MILLIS: {
                    return this.plusDays(n / 86400000L).plusNanos(n % 86400000L * 1000000L);
                }
                case SECONDS: {
                    return this.plusSeconds(n);
                }
                case MINUTES: {
                    return this.plusMinutes(n);
                }
                case HOURS: {
                    return this.plusHours(n);
                }
                case HALF_DAYS: {
                    return this.plusDays(n / 256L).plusHours(n % 256L * 12L);
                }
                default: {
                    return this.with(this.date.plus(n, temporalUnit), this.time);
                }
            }
        }
        else {
            return (LocalDateTime)temporalUnit.addTo(this, n);
        }
    }
    
    public LocalDateTime plusYears(long n) {
        return this.with(this.date.plusYears(n), this.time);
    }
    
    public LocalDateTime plusMonths(long n) {
        return this.with(this.date.plusMonths(n), this.time);
    }
    
    public LocalDateTime plusWeeks(long n) {
        return this.with(this.date.plusWeeks(n), this.time);
    }
    
    public LocalDateTime plusDays(long n) {
        return this.with(this.date.plusDays(n), this.time);
    }
    
    public LocalDateTime plusHours(long n) {
        return this.plusWithOverflow(this.date, n, 0L, 0L, 0L, 1);
    }
    
    public LocalDateTime plusMinutes(long n) {
        return this.plusWithOverflow(this.date, 0L, n, 0L, 0L, 1);
    }
    
    public LocalDateTime plusSeconds(long n) {
        return this.plusWithOverflow(this.date, 0L, 0L, n, 0L, 1);
    }
    
    public LocalDateTime plusNanos(long n) {
        return this.plusWithOverflow(this.date, 0L, 0L, 0L, n, 1);
    }
    
    @Override
    public LocalDateTime minus(TemporalAmount obj) {
        if ((obj instanceof Period)) {
            return this.with(this.date.minus((TemporalAmount)obj), this.time);
        }
        Objects.requireNonNull(obj, "amountToSubtract");
        return (LocalDateTime)obj.subtractFrom(this);
    }
    
    @Override
    public LocalDateTime minus(long n, TemporalUnit temporalUnit) {
        return (n == Long.MIN_VALUE) ? this.plus(Long.MAX_VALUE, temporalUnit).plus(1L, temporalUnit) : this.plus(-n, temporalUnit);
    }
    
    public LocalDateTime minusYears(long n) {
        return (n == Long.MIN_VALUE) ? this.plusYears(Long.MAX_VALUE).plusYears(1L) : this.plusYears(-n);
    }
    
    public LocalDateTime minusMonths(long n) {
        return (n == Long.MIN_VALUE) ? this.plusMonths(Long.MAX_VALUE).plusMonths(1L) : this.plusMonths(-n);
    }
    
    public LocalDateTime minusWeeks(long n) {
        return (n == Long.MIN_VALUE) ? this.plusWeeks(Long.MAX_VALUE).plusWeeks(1L) : this.plusWeeks(-n);
    }
    
    public LocalDateTime minusDays(long n) {
        return (n == Long.MIN_VALUE) ? this.plusDays(Long.MAX_VALUE).plusDays(1L) : this.plusDays(-n);
    }
    
    public LocalDateTime minusHours(long n) {
        return this.plusWithOverflow(this.date, n, 0L, 0L, 0L, -1);
    }
    
    public LocalDateTime minusMinutes(long n) {
        return this.plusWithOverflow(this.date, 0L, n, 0L, 0L, -1);
    }
    
    public LocalDateTime minusSeconds(long n) {
        return this.plusWithOverflow(this.date, 0L, 0L, n, 0L, -1);
    }
    
    public LocalDateTime minusNanos(long n) {
        return this.plusWithOverflow(this.date, 0L, 0L, 0L, n, -1);
    }
    
    private LocalDateTime plusWithOverflow(LocalDate localDate, long n, long n2, long n3, long n4, int n5) {
        if ((n | n2 | n3 | n4) == 0x0L) {
            return this.with(localDate, this.time);
        }
        long n6 = (n4 / 86400000000000L + n3 / 86400L + n2 / 1440L + n / 24L) * (long)n5;
        long n7 = n4 % 86400000000000L + n3 % 86400L * 1000000000L + n2 % 1440L * 60000000000L + n % 24L * 3600000000000L;
        long nanoOfDay = this.time.toNanoOfDay();
        long n8 = n7 * (long)n5 + nanoOfDay;
        long n9 = n6 + Math.floorDiv(n8, 86400000000000L);
        long floorMod = Math.floorMod(n8, 86400000000000L);
        return this.with(localDate.plusDays(n9), (floorMod == nanoOfDay) ? this.time : LocalTime.ofNanoOfDay(floorMod));
    }
    
    @Override
    public <R> R query(TemporalQuery<R> query) {
        if (query == TemporalQueries.localDate()) {
            return this.date;
        }
        return super.query(query);
    }
    
    @Override
    public Temporal adjustInto(Temporal temporal) {
        return super.adjustInto(temporal);
    }
    
    @Override
    public long until(Temporal temporal, TemporalUnit temporalUnit) {
        LocalDateTime from = from(temporal);
        if ((temporalUnit instanceof ChronoUnit)) {
            if (temporalUnit.isTimeBased()) {
                long daysUntil = this.date.daysUntil(from.date);
                if (daysUntil == 0L) {
                    return this.time.until(from.time, temporalUnit);
                }
                long n = from.time.toNanoOfDay() - this.time.toNanoOfDay();
                long n2;
                long y;
                if (daysUntil > 0L) {
                    n2 = daysUntil - 1L;
                    y = n + 86400000000000L;
                }
                else {
                    n2 = daysUntil + 1L;
                    y = n - 86400000000000L;
                }
                switch ((ChronoUnit)temporalUnit) {
                    case NANOS: {
                        n2 = Math.multiplyExact(n2, 86400000000000L);
                        break;
                    }
                    case MICROS: {
                        n2 = Math.multiplyExact(n2, 86400000000L);
                        y /= 1000L;
                        break;
                    }
                    case MILLIS: {
                        n2 = Math.multiplyExact(n2, 86400000L);
                        y /= 1000000L;
                        break;
                    }
                    case SECONDS: {
                        n2 = Math.multiplyExact(n2, 86400L);
                        y /= 1000000000L;
                        break;
                    }
                    case MINUTES: {
                        n2 = Math.multiplyExact(n2, 1440L);
                        y /= 60000000000L;
                        break;
                    }
                    case HOURS: {
                        n2 = Math.multiplyExact(n2, 24L);
                        y /= 3600000000000L;
                        break;
                    }
                    case HALF_DAYS: {
                        n2 = Math.multiplyExact(n2, 2L);
                        y /= 43200000000000L;
                        break;
                    }
                }
                return Math.addExact(n2, y);
            }
            else {
                LocalDate localDate = from.date;
                if (localDate.isAfter(this.date) && from.time.isBefore(this.time)) {
                    localDate = localDate.minusDays(1L);
                }
                else if (localDate.isBefore(this.date) && from.time.isAfter(this.time)) {
                    localDate = localDate.plusDays(1L);
                }
                return this.date.until(localDate, temporalUnit);
            }
        }
        else {
            return temporalUnit.between(this, from);
        }
    }
    
    @Override
    public String format(DateTimeFormatter obj) {
        Objects.requireNonNull(obj, "formatter");
        return obj.format(this);
    }
    
    public OffsetDateTime atOffset(ZoneOffset zoneOffset) {
        return OffsetDateTime.of(this, zoneOffset);
    }
    
    @Override
    public ZonedDateTime atZone(ZoneId zoneId) {
        return ZonedDateTime.of(this, zoneId);
    }
    
    @Override
    public int compareTo(ChronoLocalDateTime<?> other) {
        if ((other instanceof LocalDateTime)) {
            return this.compareTo0((LocalDateTime)other);
        }
        return super.compareTo(other);
    }
    
    private int compareTo0(LocalDateTime localDateTime) {
        int n = this.date.compareTo0(localDateTime.toLocalDate());
        if (n == 0) {
            n = this.time.compareTo(localDateTime.toLocalTime());
        }
        return n;
    }
    
    @Override
    public boolean isAfter(ChronoLocalDateTime<?> other) {
        if ((other instanceof LocalDateTime)) {
            return this.compareTo0((LocalDateTime)other) > 0;
        }
        return super.isAfter(other);
    }
    
    @Override
    public boolean isBefore(ChronoLocalDateTime<?> other) {
        if ((other instanceof LocalDateTime)) {
            return this.compareTo0((LocalDateTime)other) < 0;
        }
        return super.isBefore(other);
    }
    
    @Override
    public boolean isEqual(ChronoLocalDateTime<?> other) {
        if ((other instanceof LocalDateTime)) {
            return this.compareTo0((LocalDateTime)other) == 0;
        }
        return super.isEqual(other);
    }
    
    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if ((o instanceof LocalDateTime)) {
            LocalDateTime localDateTime = (LocalDateTime)o;
            return this.date.equals(localDateTime.date) && this.time.equals(localDateTime.time);
        }
        return false;
    }
    
    @Override
    public int hashCode() {
        return this.date.hashCode() ^ this.time.hashCode();
    }
    
    @Override
    public String toString() {
        return this.date.toString() + 84 + this.time.toString();
    }
    
    private Object writeReplace() {
        return new Ser((byte)5, this);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws InvalidObjectException {
        throw new InvalidObjectException("Deserialization via serialization delegate");
    }
    
    void writeExternal(DataOutput dataOutput) throws IOException {
        this.date.writeExternal(dataOutput);
        this.time.writeExternal(dataOutput);
    }
    
    static LocalDateTime readExternal(DataInput dataInput) throws IOException {
        return of(LocalDate.readExternal(dataInput), LocalTime.readExternal(dataInput));
    }
    
    static {
        MIN = of(LocalDate.MIN, LocalTime.MIN);
        MAX = of(LocalDate.MAX, LocalTime.MAX);
    }
}
